import logging
import os
from typing import Optional, Tuple

import click

from vpype import (
    CONFIG_MANAGER,
    PAGE_SIZES,
    Document,
    convert_page_size,
    global_processor,
    write_hpgl,
    write_svg,
)

from .cli import cli

__all__ = ("write",)

WRITE_HELP = f"""Save geometries to a file.

The `write` command support two format: SVG and HPGL. The format is determined based on the
file extension used for `OUTPUT` or the `--file-format` option. The latter is useful when
`OUTPUT` is a single dash (`-`), in which case the output is printed to stdout instead of a
file.

When writing to SVG, the current page size is used if available. The current page size is
implicitly set by the `read` command and can also be manually changed using the `pagesize`
command. The page size can be overridden using the `--page-size SIZE` option. `SIZE` may
be one of:

    {', '.join(PAGE_SIZES.keys())}

Alternatively, a custom size can be specified in the form of `WIDTHxHEIGHT`. `WIDTH` and
`HEIGHT` may include units. If only one has an unit, the other is assumed to have the
same unit. If none have units, both are assumed to be pixels by default. Here are some
examples:

\b
    --page-size 11x14in     # 11in by 14in
    --page-size 1024x768    # 1024px by 768px
    --page-size 13.5inx4cm  # 13.5in by 4cm

When a page size is provided, it will be rotated if the `--landscape` option is used.

If the current page set has not been set (e.g. because the `read` command has not been used)
and the `--page-size` is not provided, the SVG will have its bounds tightly fit to the
geometries.

By default, the geometries are not scaled or translated even if they lie outside of the page
boundaries. The `--center` option translates the geometries to the center of the page.

Layers are labelled with their numbers by default. If an alternative naming is required, a
template pattern can be provided using the `--layer-label` option. The provided pattern must
contain a C-style format specifier such as `%d` which will be replaced by the layer number.

For previsualization purposes, paths are colored by layer in the SVG. This can be controlled
with the `--color-mode` option. Setting it "none" disables coloring and black paths are
generated. Setting it to "path" gives a different color to each path (with a rotation),
which makes it easier to visualize line optimization. Finally, pen-up trajectories can be
generated with the `--pen-up` flag. As most plotting tools will include these paths in the
output, this option should be used for previsualisation only. The Axidraw tools will however
ignore them.

When writing to HPGL, a device name must be provided with the `--device` option. The
corresponding device must be configured in the built-in or a user-provided configuration file
(see the documentation for more details). The following devices are currently available:

    {', '.join(CONFIG_MANAGER.get_plotter_list())}

In HPGL mode, because the coordinate system depends on the configuration, the `--page-size`
option is mandatory, and is restricted to the paper formats defined in the corresponding
device's configuration. The plotter may as well need to be specifically configured for the
desired paper size (e.g. for A4 or A3, the HP 7475a's corresponding DIP switch must be set to
metric mode).

The `--landscape` and `--center` options are accepted and honored in HPGL.

Optionally, the HPGL-only `--velocity` can be provided, in which case a `VS` command will be
emitted with the provided value.

Examples:

    Write to a SVG using the current page size as set by the `read` command:

        vpype read input.svg [...] write output.svg

    Write to a portrait A4 page:

        vpype [...] write --page-size a4 output.svg

    Write to a 13x9 inch page and center the geometries:

        vpype [...] write --page-size 13x9in --landscape --center output.svg

    Use custom layer labels:

        vpype [...] write --page-size a4 --layer-label "Pen %d" output.svg

    Write a previsualization SVG:

        vpype [...] write --pen-up --color-mode path output.svg

    Output SVG to stdout:

        vpype [...] write --format SVG -

    Write a A4 page with portrait orientation HPGL file:

        vpype [...] write --device hp7475a --page-size a4 --center
"""


@cli.command(group="Output", help=WRITE_HELP)
@click.argument("output", type=click.File("w"))
@click.option(
    "-f",
    "--format",
    "file_format",
    type=click.Choice(["svg", "hpgl"], case_sensitive=False),
    help="Output format (inferred from file extension by default).",
)
@click.option(
    "-p",
    "--page-size",
    type=str,
    help=(
        "Set the bounds of the SVG to a specific page size. If omitted, the SVG size is set "
        "to the current page size (see `read` and `pagesize` commands. May not be omitted for "
        "HPGL."
    ),
)
@click.option(
    "-l", "--landscape", is_flag=True, help="Use landscape orientation instead of portrait."
)
@click.option(
    "-c", "--center", is_flag=True, help="Center the geometries within the SVG bounds."
)
@click.option(
    "-ll",
    "--layer-label",
    type=str,
    default="%d",
    help="[SVG only] Pattern used to for naming layers.",
)
@click.option("-pu", "--pen-up", is_flag=True, help="[SVG only] Generate pen-up trajectories.")
@click.option(
    "-m",
    "--color-mode",
    type=click.Choice(["none", "layer", "path"]),
    default="layer",
    help="[SVG only] Color mode for paths (default: layer).",
)
@click.option("-d", "--device", type=str, help="[HPGL only] Type of the plotter device.")
@click.option(
    "-vs",
    "--velocity",
    type=float,
    help="[HPGL only] Emit a VS command with the provided value.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="[HPGL only] Do not display the plotter configuration or paper loading information.",
)
@click.pass_obj  # to obtain the command string
@global_processor
def write(
    document: Document,
    cmd_string: Optional[str],
    output,
    file_format: str,
    page_size: str,
    landscape: bool,
    center: bool,
    layer_label: str,
    pen_up: bool,
    color_mode: str,
    device: Optional[str],
    velocity: Optional[int],
    quiet: bool,
):
    """Write command."""

    if document.is_empty():
        logging.warning("no geometry to save, no file created")
        return document

    if file_format is None:
        # infer format
        _, ext = os.path.splitext(output.name)
        file_format = ext.lstrip(".").lower()

    if file_format == "svg":
        page_size_px = None
        if page_size:
            page_size_px = convert_page_size(page_size)
            if landscape:
                page_size_px = page_size_px[::-1]

        write_svg(
            output=output,
            document=document,
            page_size=page_size_px,
            center=center,
            source_string=cmd_string if cmd_string is not None else "",
            layer_label_format=layer_label,
            show_pen_up=pen_up,
            color_mode=color_mode,
        )
    elif file_format == "hpgl":
        if page_size:
            write_hpgl(
                output=output,
                document=document,
                landscape=landscape,
                center=center,
                device=device,
                page_size=page_size,
                velocity=velocity,
                quiet=quiet,
            )
        else:
            logging.error(
                "write: page size is mandatory for HPGL, please use the `--page-size SIZE` "
                "option"
            )
    else:
        logging.warning(
            f"write: format could not be inferred or format unknown '{file_format}', "
            "no file created"
        )

    return document



import logging
import math
import re
from typing import Callable, Dict, List, Tuple, Union

import click
import numpy as np

# REMINDER: anything added here must be added to docs/api.rst
__all__ = [
    "UNITS",
    "ANGLE_UNITS",
    "PAGE_SIZES",
    "LengthType",
    "AngleType",
    "PageSizeType",
    "convert_length",
    "convert_angle",
    "convert_page_size",
    "union",
    # deprecated:
    "PAGE_FORMATS",
    "convert",
    "convert_page_format",
    "Length",
]


def _mm_to_px(x: float, y: float) -> Tuple[float, float]:
    return x * 96.0 / 25.4, y * 96.0 / 25.4


UNITS = {
    "px": 1.0,
    "in": 96.0,
    "mm": 96.0 / 25.4,
    "cm": 96.0 / 2.54,
    "pc": 16.0,
    "pt": 96.0 / 72.0,
}

ANGLE_UNITS = {
    "deg": 1.0,
    "grad": 9.0 / 10.0,  # note: must be before "rad"!
    "rad": 180.0 / math.pi,
    "turn": 360.0,
}

# page sizes in pixel
PAGE_SIZES = {
    "tight": _mm_to_px(0, 0),
    "a6": _mm_to_px(105.0, 148.0),
    "a5": _mm_to_px(148.0, 210.0),
    "a4": _mm_to_px(210.0, 297.0),
    "a3": _mm_to_px(297.0, 420.0),
    "letter": _mm_to_px(215.9, 279.4),
    "legal": _mm_to_px(215.9, 355.6),
    "executive": _mm_to_px(185.15, 266.7),
    "tabloid": _mm_to_px(279.4, 431.8),
}

# deprecated
PAGE_FORMATS = PAGE_SIZES


def _convert_unit(value: Union[str, float], units: Dict[str, float]) -> float:
    """Converts a string with unit to a value"""
    if isinstance(value, str):
        value = value.strip().lower()
        for unit, factor in units.items():
            if value.endswith(unit):
                num = value.strip(unit)
                return (float(num) if len(num) > 0 else 1.0) * factor

    return float(value)


def convert_length(value: Union[str, float]) -> float:
    """Convert a length optionally expressed as a string with unit to px value.

    Args:
        value: value to convert

    Returns:
        converted value

    Raises:
        :class:`ValueError`
    """
    return _convert_unit(value, UNITS)


def convert(value: Union[str, float]) -> float:  # pragma: no cover
    """Deprecated, use convert_length."""
    logging.warning(
        "!!! `vpype.convert()` is deprecated, use `vpype.convert_length()` instead."
    )
    return convert_length(value)


def convert_angle(value: Union[str, float]) -> float:
    """Convert an angle optionally expressed as a string with unit to degrees.

    Args:
        value: angle to convert

    Returns:
        converted angle in degree

    Raises:
        :class:`ValueError`
    """
    return _convert_unit(value, ANGLE_UNITS)


def convert_page_size(value: str) -> Tuple[float, float]:
    """Converts a string with page size to dimension in pixels.

    The input can be either a known page size (see ``vpype write --help`` for a list) or
    a page size descriptor in the form of "WxH" where both W and H can have units.

    Examples:

        Using a know page size::

            >>> import vpype
            >>> vpype.convert_page_size("a3")
            (1122.5196850393702, 1587.4015748031497)

        Using page size descriptor (no units, pixels are assumed)::

            >>> vpype.convert_page_size("100x200")
            (100.0, 200.0)

        Using page size descriptor (explicit units)::

            >>> vpype.convert_page_size("1inx2in")
            (96.0, 192.0)

    Args:
        value: page size descriptor

    Returns:
        the page size in CSS pixels
    """
    if value in PAGE_SIZES:
        return PAGE_SIZES[value]

    match = re.match(
        r"^(\d+\.?\d*)({0})?x(\d+\.?\d*)({0})?$".format("|".join(UNITS.keys())), value
    )

    if not match:
        raise ValueError(f"page size '{value}' unknown")

    x, x_unit, y, y_unit = match.groups()

    if not x_unit:
        x_unit = y_unit if y_unit else "px"
    if not y_unit:
        y_unit = x_unit

    return float(x) * convert_length(x_unit), float(y) * convert_length(y_unit)


def convert_page_format(value: str) -> Tuple[float, float]:  # pragma: no cover
    """Deprecated, use convert_page_size."""
    logging.warning(
        "!!! `vpype.convert_page_format()` is deprecated, use `vpype.convert_page_size()` "
        "instead."
    )
    return convert_page_size(value)


class LengthType(click.ParamType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided length
    string (which may contain units) into a value in CSS pixel units. This class uses
    :func:`convert_length` internally.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("x", type=vpype.LengthType())
        ... @click.option("-o", "--option", type=vpype.LengthType(), default="1mm")
        ... @vpype.generator
        ... def my_command(x: float, option: float):
        ...     pass
    """

    name = "length"

    def convert(self, value, param, ctx):
        try:
            return convert_length(value)
        except ValueError:
            self.fail(f"parameter {value} is an incorrect length")


class Length(LengthType):  # pragma: no cover
    """Deprecated, use LengthType."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        logging.warning("!!! `vpype.Length` is deprecated, use `vpype.LengthType` instead.")


class AngleType(click.ParamType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided angle
    string (which may contain units) into a value in degrees. This class uses
    :func:`convert_angle` internally.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("angle", type=vpype.AngleType())
        ... @vpype.generator
        ... def my_command(angle: float):
        ...     pass
    """

    name = "angle"

    def convert(self, value, param, ctx):
        try:
            return convert_angle(value)
        except ValueError:
            self.fail(f"parameter {value} is an incorrect angle")


class PageSizeType(click.ParamType):
    """:class:`click.ParamType` sub-class to automatically converts a user-provided page size
    string into a tuple of float in CSS pixel units. See :func:`convert_page_size` for
    information on the page size descriptor syntax.

    Example::

        >>> import click
        >>> import vpype_cli
        >>> import vpype
        >>> @vpype_cli.cli.command(group="my commands")
        ... @click.argument("fmt", type=vpype.PageSizeType())
        ... @vpype.generator
        ... def my_command(fmt: Tuple[float, float]):
        ...     pass
    """

    name = "PAGESIZE"

    def convert(self, value, param, ctx) -> Tuple[float, float]:
        try:
            return convert_page_size(value)
        except ValueError:
            self.fail(f"parameter {value} is not a valid page size")


def union(line: np.ndarray, keys: List[Callable[[np.ndarray], bool]]) -> bool:
    """Returns True if every callables in ``keys`` return True (similar to ``all()``. This
    function is typically used with :meth:`LineCollection.filter`.

    Args:
        line: line to test
        keys: list of callables

    Returns:
        True if every callables return True
    """
    for key in keys:
        if not key(line):
            return False
    return True



import click

from vpype import (
    Document,
    LayerType,
    LineCollection,
    global_processor,
    multiple_to_layer_ids,
    single_to_layer_id,
)

from .cli import cli

__all__ = ("lcopy", "lmove", "ldelete")


@cli.command(group="Layers")
@click.argument("sources", type=LayerType(accept_multiple=True))
@click.argument("dest", type=LayerType(accept_new=True))
@global_processor
def lcopy(document, sources, dest):
    """Copy the content of one or more layer(s) to another layer.

    SOURCES can be a single layer ID, the string 'all' (to copy all non-empty layers,
    or a coma-separated, whitespace-free list of layer IDs.

    DEST can be a layer ID or the string 'new', in which case a new layer with the
    lowest available ID will be created.

    If a layer is both in the source and destination, its content is not duplicated.

    Examples:

        Copy layer 1 to a new layer:

            vpype [...] lcopy 1 new [...]  # duplicate layer 1

        Make a new layer with a merged copy of layer 1 and 2:

            vpype [...] lcopy 1,2 new [...]  # make new layer with merged copy of layer 1 and 2

        Add a merged copy of all layers to layer 1. If layer 1 previously had content, this \
content is not duplicated:

            vpype [...] lcopy all 1 [...]
    """

    src_lids = multiple_to_layer_ids(sources, document)
    dest_lid = single_to_layer_id(dest, document)

    if dest_lid in src_lids:
        src_lids.remove(dest_lid)

    lc = LineCollection()
    for lid in src_lids:
        lc.extend(document[lid])
    document.add(lc, dest_lid)

    return document


@cli.command(group="Layers")
@click.argument("sources", type=LayerType(accept_multiple=True))
@click.argument("dest", type=LayerType(accept_new=True))
@global_processor
def lmove(document, sources, dest):
    """Move the content of one or more layer(s) to another layer.

    SOURCES can be a single layer ID, the string 'all' (to copy all non-empty layers,
    or a coma-separated, whitespace-free list of layer IDs.

    DEST can be a layer ID or the string 'new', in which case a new layer with the
    lowest available ID will be created.

    Layer(s) left empty after the move are then discarded and may thus be reused by subsequent
    commands using 'new' as destination layer.

    If a layer is both in the source and destination, its content is not duplicated.

    Examples:

        Merge layer 1 and 2 to layer 1 (the content of layer 1 is not duplicated):

            vpype [...] lmove 1,2 1 [...]  # merge layer 1 and 2 to layer 1
    """

    src_lids = multiple_to_layer_ids(sources, document)
    dest_lid = single_to_layer_id(dest, document)

    if dest_lid in src_lids:
        src_lids.remove(dest_lid)

    for lid in src_lids:
        document.add(document.pop(lid), dest_lid)

    return document


@cli.command(group="Layers")
@click.argument("layers", type=LayerType(accept_multiple=True))
@global_processor
def ldelete(document: Document, layers) -> Document:
    """Delete one or more layers.

    LAYERS can be a single layer ID, the string 'all' (to delete all layers), or a
    coma-separated, whitespace-free list of layer IDs.
    """

    lids = set(multiple_to_layer_ids(layers, document))

    new_doc = document.empty_copy()
    for lid in document.ids():
        if lid not in lids:
            new_doc[lid] = document[lid]

    return new_doc
